"use strict";
/* tslint:disable no-console */
Object.defineProperty(exports, "__esModule", { value: true });
exports.generate = exports.LogLevel = void 0;
const chalk = require("chalk");
const json_schema_to_typescript_1 = require("json-schema-to-typescript");
const lodash_1 = require("lodash");
const path = require("path");
const files_1 = require("./files");
var LogLevel;
(function (LogLevel) {
    LogLevel[LogLevel["NONE"] = 0] = "NONE";
    LogLevel[LogLevel["ERROR"] = 1] = "ERROR";
    LogLevel[LogLevel["INFO"] = 2] = "INFO";
})(LogLevel = exports.LogLevel || (exports.LogLevel = {}));
let baseOptions = {
    verbosity: LogLevel.NONE,
    bannerComment: '',
    cwd: process.cwd(),
    declareExternallyReferenced: false,
    enableConstEnums: false,
    ignoreMinAndMaxItems: false,
    unknownAny: false,
    unreachableDefinitions: false,
    strictIndexSignatures: false,
    style: {
        bracketSpacing: true,
        printWidth: 120,
        semi: true,
        singleQuote: true,
        tabWidth: 4,
        useTabs: false,
    },
    $refOptions: {},
};
function logInfo(message) {
    if (baseOptions.verbosity >= LogLevel.INFO) {
        console.log(chalk.cyan.bold(message), 'SpecParser');
    }
}
function logSuccess(message) {
    if (baseOptions.verbosity >= LogLevel.INFO) {
        console.log(chalk.green.bold(message), 'SpecParser');
    }
}
function logError(message) {
    if (baseOptions.verbosity >= LogLevel.ERROR) {
        console.log(chalk.red.bold(message), 'SpecParser');
    }
}
function extractRefsFromSchema(inputSchema) {
    const objectSchema = inputSchema;
    const refSchema = inputSchema;
    switch (objectSchema.type) {
        case 'object':
            return Object.values(preprocessProperties(objectSchema.properties));
        case 'array':
            return extractRefsFromSchema(objectSchema.items);
        default:
            if (objectSchema.oneOf) {
                return (0, lodash_1.flatten)(Object.values(objectSchema.oneOf).map((item) => extractRefsFromSchema(item)));
            }
            if (objectSchema.anyOf) {
                return (0, lodash_1.flatten)(Object.values(objectSchema.anyOf).map((item) => extractRefsFromSchema(item)));
            }
            if (objectSchema.allOf) {
                return (0, lodash_1.flatten)(Object.values(objectSchema.anyOf).map((item) => extractRefsFromSchema(item)));
            }
            return refSchema.$ref || undefined;
    }
}
function preprocessProperties(inputProperties) {
    if (inputProperties) {
        return Object.entries(inputProperties).reduce((acc, [name, value]) => (Object.assign(Object.assign({}, acc), { [name]: extractRefsFromSchema(value) })), {});
    }
    return {};
}
async function createImport(name, schema, filePath) {
    const references = extractRefsFromSchema(schema);
    if (!references || !references.length) {
        return [];
    }
    const refsArray = typeof references === 'string' ? [references] : references.flat();
    const $refs = refsArray.filter((val, index) => val !== undefined && refsArray.indexOf(val) === index).map((val) => val.split('/')[3]);
    const imports = $refs.filter((val) => !val.includes(name));
    if ($refs.length && imports.length) {
        await (0, files_1.appendFile)(filePath, `import { ${imports.join(', ')} } from './'; \n\n`);
    }
    return $refs;
}
async function createInterfaceFile(name, interfacesDirPath) {
    const bannerComment = '/* tslint:disable */\n/**\n* This file was automatically generated.\n* DO NOT MODIFY IT BY HAND.\n*/\n\n';
    const filePath = `${interfacesDirPath}/${name}.ts`;
    await (0, files_1.writeFile)(filePath, bannerComment);
    return filePath;
}
function removeEnum(content) {
    const enumStartPosition = content.search('export enum');
    const interfaceStartPosition = content.search('export interface');
    if (enumStartPosition > -1 && interfaceStartPosition > -1) {
        const enumEndPosition = content.slice(enumStartPosition, content.length).indexOf('}') + enumStartPosition;
        const enumParts = content.slice(enumStartPosition, enumEndPosition + 1);
        content = content.replace(enumParts, '').trim();
        return removeEnum(content);
    }
    return content;
}
function checkSelfReference(name, content) {
    const interfaceStartPosition = content.search('export interface');
    const interfaceEndPosition = content.slice(interfaceStartPosition, content.length).indexOf('}') + interfaceStartPosition;
    const interfaceParts = content.slice(interfaceStartPosition, interfaceEndPosition + 1).split('{');
    const selfRefIndex = interfaceParts[1].indexOf(name);
    if (selfRefIndex > -1) {
        const selfRef = interfaceParts[1].slice(selfRefIndex, selfRefIndex + name.length + 1);
        const regex = new RegExp(selfRef, 'g');
        content = content.replace(regex, name);
    }
    return content;
}
async function createInterfaceContent(name, openApiSpec, filePath) {
    const components = openApiSpec.components;
    const schema = openApiSpec.components.schemas[name];
    const dependencies = await createImport(name, schema, filePath);
    const schemaAndDefinitions = Object.assign(Object.assign({}, schema), { components });
    let content = await (0, json_schema_to_typescript_1.compile)(schemaAndDefinitions, name, baseOptions);
    // TODO: check future updates from json-schema-to-typescript, as next step should not be needed
    if (!baseOptions.declareExternallyReferenced) {
        content = removeEnum(content);
    }
    if (dependencies.includes(name)) {
        content = checkSelfReference(name, content);
    }
    await (0, files_1.appendFile)(filePath, content);
    return content;
}
async function createInterface(name, openApiSpec, interfacesDirPath) {
    await (0, files_1.appendFile)(`${interfacesDirPath}/index.ts`, `export * from './${name}';\n`);
    const filePath = await createInterfaceFile(name, interfacesDirPath);
    const content = await createInterfaceContent(name, openApiSpec, filePath);
    logInfo(`${name} interface generated`);
    return { content, filePath, name };
}
async function parseSchema(name, openApiSpec, interfacesDirPath) {
    const components = openApiSpec === null || openApiSpec === void 0 ? void 0 : openApiSpec.components;
    if (!components || !components.schemas || !components.schemas.hasOwnProperty(name)) {
        return null;
    }
    const schema = components.schemas[name];
    return schema ? createInterface(name, openApiSpec, interfacesDirPath) : null;
}
function delayedParsing(schemaKey, openApiSpec, interfacesDirPath, ms) {
    return new Promise((resolve, reject) => setTimeout(() => parseSchema(schemaKey, openApiSpec, interfacesDirPath)
        .then((res) => resolve(res))
        .catch((e) => reject(e)), ms));
}
async function openApiToInterfaces(openApiSpec, interfacesDirPath) {
    const schemasNames = Object.keys(openApiSpec.components.schemas).sort();
    await schemasNames.reduce(async (prevPromise, schemaKey) => {
        try {
            await prevPromise;
            return delayedParsing(schemaKey, openApiSpec, interfacesDirPath, 0);
        }
        catch (error) {
            logError(error.message);
            return Promise.resolve({});
        }
    }, Promise.resolve({}));
}
async function removeExistingInterfaces(interfacesPath) {
    const files = await (0, files_1.readDir)(interfacesPath);
    return Promise.all(Object.values(files).map((file) => (0, files_1.removeFile)(path.join(interfacesPath, file))));
}
function appendTitles(openApiSpec) {
    Object.entries(openApiSpec.components.schemas).forEach(([name, schema]) => {
        const objectSchema = schema;
        if (objectSchema.type) {
            objectSchema.title = objectSchema.title || name;
            if (objectSchema.type === 'string' && objectSchema.enum && !objectSchema.tsEnumNames) {
                objectSchema.tsEnumNames = objectSchema.enum.map((e) => (0, lodash_1.snakeCase)(e).toUpperCase());
            }
        }
    });
}
async function generate(openApiFilePath = './openapi.json', interfacesDirPath = './interfaces', options = {}) {
    baseOptions = Object.assign(Object.assign({}, baseOptions), options);
    try {
        const swaggerSpec = JSON.parse(await (0, files_1.readFile)(openApiFilePath));
        appendTitles(swaggerSpec);
        await removeExistingInterfaces(interfacesDirPath);
        await openApiToInterfaces(swaggerSpec, interfacesDirPath);
        logSuccess('Typescript interfaces generated');
    }
    catch (error) {
        logError(error.message);
        throw error;
    }
}
exports.generate = generate;
//# sourceMappingURL=spec-parser.js.map